%{  open Ast 
    open Lexing
    let parse_error msg = 
        let start_pos = Parsing.rhs_start_pos 1 in
				        let lineNo = start_pos.pos_lnum in
				        print_endline (msg ^ " in line #" ^ string_of_int lineNo)
%}

%token SEMI LPAREN RPAREN LBRACE RBRACE COMMA COLON LBRACKET RBRACKET NEWMAT
%token PLUS MINUS MUL DIVIDE ASSIGN  POWER EMUL ERDIV LDIV ELDIV TRANSPOSE
%token EQ NOTEQ LT LTEQ GT GTEQ AND OR NOT
%token IF ELSE ELSEIF FOR WHILE BREAK CONTINUE END FUNCTION
%token <int> LITERAL
%token <float> FLOATLIT
%token <string> ID
%token <string> DATATYPE
%token <string> STRING
%token <string> BOOL
%token EOF

%nonassoc NOELSE
%nonassoc ELSE

%nonassoc NOACTUALS
%nonassoc LPAREN

%nonassoc NOMINUS

%right ASSIGN
%left OR
%left AND
%left EQ NOTEQ
%left LT GT LTEQ GTEQ
%left PLUS MINUS
%left MUL DIVIDE EMUL LDIV ELDIV ERDIV
%right POWER
%left TRANSPOSE
%right NOT

%nonassoc UMINUS

%start program
%type <Ast.program> program

%%

program:
   /* nothing */ { [], [] }
 | program vdecl { ($2 :: fst $1), snd $1 }
 | program fdecl { fst $1, ($2 :: snd $1) }

fdecl:
   FUNCTION DATATYPE ID ASSIGN ID LPAREN formals_opt RPAREN vdecl_list stmt_list END
     { { fname = $5;
     rettype = $2;
     retname = $3;
	 formals = $7;
	 locals = List.rev $9;
	 body = List.rev $10 } } 
  |  FUNCTION ID LPAREN formals_opt RPAREN vdecl_list stmt_list END
     { { fname = $2;
     rettype = "void";
     retname = "";
	 formals = $4;
	 locals = List.rev $6;
	 body = List.rev $7 } }

formals_opt:
    /* nothing */ { [] }
  | formal_list   { List.rev $1 }

formal_list:
    param_decl          { [$1] }
  | formal_list COMMA param_decl  { $3 :: $1 }
  
param_decl:
   DATATYPE ID 
     { { varname = $2; 
     vartype = $1 } }

vdecl_list:
    /* nothing */    { [] }
  | vdecl_list vdecl { $2 :: $1 }

vdecl:
   DATATYPE ID SEMI
     { { varname = $2; 
     vartype = $1 } }

	    
stmt_list:
    /* nothing */ %prec NOMINUS { [] }
  | stmt_list stmt { $2 :: $1 }
  


stmt:
    expr SEMI { Expr($1) }
  | error SEMI { Expr(Datatype("parseerror")) } 
  | BREAK SEMI { Break }
  | CONTINUE SEMI { Continue }
  | IF expr stmt_list ELSE stmt_list END 
        { If($2, Block(List.rev $3), Block(List.rev $5)) } 
  | IF expr stmt_list elseif_list ELSE stmt_list END 
        { Ifelseif($2, Block(List.rev $3), List.rev $4, Block(List.rev $6)) }
  | FOR for_expr stmt_list END
        { For($2, Block(List.rev $3)) }
  | WHILE expr stmt_list END { While($2, Block(List.rev $3)) }
  | ID ASSIGN LBRACKET matrix RBRACKET SEMI
        { MatrixInit($1, List.rev $4) }

matrix: /*matrix definition*/
    expr          { [[$1]] }
  | matrix COMMA expr  { 
        if (List.length $1) = 1 then
            [$3 :: List.hd $1] 
        else
            ($3 :: List.hd $1) :: List.tl $1 }
  | matrix SEMI expr  { [$3] :: $1 }


expr:
    LITERAL          { Literal($1) }
  | FLOATLIT         { Floatlit($1) }
  | BOOL             { Bool($1) }
  | ID %prec NOACTUALS     { Id($1) }
  | STRING           { String($1) }
  | expr PLUS   expr { Binop($1, Add,   $3) }
  | expr MINUS  expr { Binop($1, Sub,   $3) }
  | expr MUL  expr { Binop($1, Mul,  $3) }
  | expr DIVIDE expr { Binop($1, Div,   $3) }
  | expr AND   expr  { Binop($1, And,   $3) }
  | expr OR  expr    { Binop($1, Or,   $3) } 
  | expr POWER  expr { Binop($1, Power,  $3) }  
  | NOT expr         { Not($2) }   
  | MINUS expr %prec UMINUS    { Uminus($2) }
  | expr TRANSPOSE {Transpose($1)}
  | expr EQ     expr { Binop($1, Equal, $3) }
  | expr NOTEQ    expr { Binop($1, Neq,   $3) }
  | expr LT     expr { Binop($1, Less,  $3) }
  | expr LTEQ    expr { Binop($1, Leq,   $3) }
  | expr GT     expr { Binop($1, Greater,  $3) }
  | expr GTEQ    expr { Binop($1, Geq,   $3) } 
  | expr EMUL    expr {Binop($1, Emul,$3)}
  | expr ERDIV  expr {Binop($1, Erdiv, $3)}
  | expr LDIV expr {Binop($1, Ldiv, $3)}
  | expr ELDIV  expr {Binop($1, Eldiv, $3)}
  | ID ASSIGN expr   { Assign($1, $3) }
  | ID LBRACKET actuals_list RBRACKET ASSIGN expr 
        {AssignMatrix($1, $3, $6)}   
  | ID LPAREN actuals_opt RPAREN { Call($1, $3) }
  | LPAREN expr RPAREN { Paren($2) }
  | ID LBRACKET actuals_list RBRACKET
        { MatrixAccess($1, $3) }
  | ID LBRACKET expr COLON expr COMMA expr COLON expr RBRACKET {MatrixRAccess ($1, $3, $5, $7, $9)}
  | NEWMAT LPAREN expr COMMA expr RPAREN
        { Newmat($3, $5) }

for_expr:
    ID ASSIGN expr COLON expr %prec NOMINUS {Assigna($1, $3, $5)}
  | ID ASSIGN expr COLON expr COLON expr %prec NOMINUS
    {Assignb($1, $3, $5, $7)}

elseif_list:
    elseif_clause             { [$1] }
  | elseif_list elseif_clause { $2 :: $1 }

elseif_clause:
    ELSEIF expr stmt_list 
     { {elseif_expr = $2; 
        elseif_stmt = Block(List.rev $3)} }

actuals_opt:
    /* nothing */ { [] }
  | actuals_list  { List.rev $1 }

actuals_list:
    expr                    { [$1] }
  | actuals_list COMMA expr { $3 :: $1 }
